/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: pam.c,v 1.5 2021/05/29 10:27:42 babolo Exp $"

#define BLIN_COMPAT     4
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <mife.h>
#include "pngtr.h"

static size_t
/*****************************************************************************************************
 ****                                        ****
 *****************************************************************************************************
                                                                                                  ****/
coord(pngtr_convey *convey, u_int r, u_int c) {                                                  /****
                                                                                                  ****
 *****************************************************************************************************/
    u_int x, y, h, w;

    if  (convey->flags & PNGTR_ROTSWAP) {
        x = r; y = c; w = convey->IHDR.height; h = convey->IHDR.width;
    } else {
        x = c; y = r; w = convey->IHDR.width; h = convey->IHDR.height;
    }
    if  (convey->flags & PNGTR_ROTATEX) x = w - x - 1;
    if  (convey->flags & PNGTR_ROTATEY) y = h - y - 1;
    return(w * y + x);
}

pngtr_syncolor *
/*****************************************************************************************************
 *****************************************************************************************************
 ****    fd  cf->num    f                                           ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_dopam(pngtr_conf *cf, u_int num) {                                                         /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_syncolor    syncolor;
    pngtr_syncolor    smcolor;
    pngtr_operand    *palette  = NULL;
    pngtr_convey     *convey;
    u_int16_t         acolor;
    pngtr_operand    *alfa     = NULL;
    u_int16_t         grey;
    pngtr_syncolor   *img      = NULL;
    int               ex       = EX_OK;
#   define O(R, C) img[coord(convey, r, c)]

    ifBLIN_QX3("+ %08X %u", cf->flags, num);
    PNGTR_X((num >= PNGTR_CONVEYNUM),                    ("Ill convey num %d", num), ERANGE, EX_USAGE);
    convey = &cf->convey[num];
    for (u_int chctr = 0; chctr < convey->chunum; ++chctr) {
        if  (PNGTR_tRNS == convey->chunks[chctr].type) {
            alfa = &convey->chunks[chctr].operands[0];
            break;
    }   }
    PNGTR_W( !(img = calloc(convey->IHDR.width * convey->IHDR.height, sizeof(pngtr_syncolor)))
           , ("calloc")
           , EX_OSERR
           );
    if  (convey->IHDR.palette) {
        for (u_int i = 0; i < convey->chunum; ++i) {
            if  (PNGTR_PLTE == convey->chunks[i].type) {
                palette = convey->chunks[i].operands;
                break;
    }   }   }
    switch(convey->IHDR.depth) {
#   define I(R, C, N) convey->img.cont[((R) * convey->IHDR.width + (N) - 1) / (N) + (C) / (N)]
    case 1:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 8) >> (7 - (c & 7))) & 1;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    if  (!!grey) grey = 65535;
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 8) >> (7 - (c & 7))) & 1;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    case 2:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 4) >> (6 - 2 * (c & 3))) & 3;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    grey = (u_int16_t[]){0x0000, 0x5555, 0xAAAA, 0xFFFF}[grey];
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 4) >> (6 - 2 * (c & 3))) & 3;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    case 4:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 2) >> (4 - 4 * (c & 1))) & 15;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    grey = (u_int16_t[]){0x0000, 0x1111, 0x2222, 0x3333, 0x4444, 0x5555, 0x6666, 0x7777, 0x8888, 0x9999, 0xAAAA, 0xBBBB, 0xCCCC, 0xDDDD, 0xEEEE, 0xFFFF}[grey];
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 2) >> (4 - 4 * (c & 1))) & 15;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    case 8:
        switch(convey->IHDR.colortype) {
#   undef  I
#   define X(R, C, N, M) ((N) * ((R) * convey->IHDR.width + (C)) + (M))
#   define I(R, C, N, M) convey->img.cont[X((R), (C), (N), (M))]
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 1, 0) & 0x00FF;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    grey |= grey << 8;
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 2:
            if  (!!alfa && (0 < alfa->maxindex)) {
                syncolor.r = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->r, -2);
                syncolor.g = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->g, -2);
                syncolor.b = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->b, -2);
            }
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    O(r, c).r = I(r, c, 3, 0) & 0x00FF;
                    O(r, c).g = I(r, c, 3, 1) & 0x00FF;
                    O(r, c).b = I(r, c, 3, 2) & 0x00FF;
                    if  (  !!alfa
                        && (0 < alfa->maxindex)
                        && (O(r, c).r == syncolor.r)
                        && (O(r, c).g == syncolor.g)
                        && (O(r, c).b == syncolor.b)
                        ) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    O(r, c).r |= O(r, c).r << 8;
                    O(r, c).g |= O(r, c).g << 8;
                    O(r, c).b |= O(r, c).b << 8;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 1, 0) & 0x00FF;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        case 4:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 2, 0) & 0x00FF;
                    grey |= grey << 8;
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
                    O(r, c).a = I(r, c, 2, 1) & 0x00FF;
                    O(r, c).a |= O(r, c).a << 8;
            }   }
            break;
        case 6:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    O(r, c).r = I(r, c, 4, 0) & 0x00FF;
                    O(r, c).g = I(r, c, 4, 1) & 0x00FF;
                    O(r, c).b = I(r, c, 4, 2) & 0x00FF;
                    O(r, c).a = I(r, c, 4, 3) & 0x00FF;
                    O(r, c).r |= O(r, c).r << 8;
                    O(r, c).g |= O(r, c).g << 8;
                    O(r, c).b |= O(r, c).b << 8;
                    O(r, c).a |= O(r, c).a << 8;
            }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
#   undef  I
#   define I(R, C, N, M) (u_int16_t)recobe_load(&convey->img.cont[2 * X((R), (C), (N), (M))], -2)
    case 16:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 1, 0);
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    recobe_store((u_char *)&O(r, c).r, -2, grey);
                    recobe_store((u_char *)&O(r, c).g, -2, grey);
                    recobe_store((u_char *)&O(r, c).b, -2, grey);
            }   }
            break;
        case 2:
            if  (!!alfa && (0 < alfa->maxindex)) {
                syncolor.r = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->r, -2);
                syncolor.g = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->g, -2);
                syncolor.b = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->b, -2);
            }
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    smcolor.r = I(r, c, 3, 0);
                    smcolor.g = I(r, c, 3, 1);
                    smcolor.b = I(r, c, 3, 2);
                    if  (  !!alfa
                        && (0 < alfa->maxindex)
                        && (smcolor.r == syncolor.r)
                        && (smcolor.g == syncolor.g)
                        && (smcolor.b == syncolor.b)
                        ) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    recobe_store((u_char *)&O(r, c).r, -2, smcolor.r);
                    recobe_store((u_char *)&O(r, c).g, -2, smcolor.g);
                    recobe_store((u_char *)&O(r, c).b, -2, smcolor.b);
            }   }
            break;
        case 4:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 2, 0);
                    recobe_store((u_char *)&O(r, c).r, -2, grey);
                    recobe_store((u_char *)&O(r, c).g, -2, grey);
                    recobe_store((u_char *)&O(r, c).b, -2, grey);
                    recobe_store((u_char *)&O(r, c).a, -2, I(r, c, 2, 1));
            }   }
            break;
        case 6:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    recobe_store((u_char *)&O(r, c).r, -2, I(r, c, 4, 0));
                    recobe_store((u_char *)&O(r, c).g, -2, I(r, c, 4, 1));
                    recobe_store((u_char *)&O(r, c).b, -2, I(r, c, 4, 2));
                    recobe_store((u_char *)&O(r, c).a, -2, I(r, c, 4, 3));
            }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    default:
        ifBLIN_QX0( "depth %u in %s"
                  , convey->IHDR.depth
                  , !convey->png.nm ? "" : convey->png.nm
                  );
        ex = EX_USAGE;
        errno = EINVAL;
        goto out;
    }
out:
    ifBLIN_QX3("- %d", ex);
    if  (!!ex) {
        free(img);
        img = NULL;
    }
    return(img);
#   undef  I
#   undef  O
#   undef  blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    fd  cf->num    f                                           ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_outpam(pngtr_conf *cf, int fd) {                                                           /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;
    ssize_t           dopam;
    const char       *pngnm;
    int               hlen;
    pngtr_syncolor   *img   = NULL;
    char             *buf;
    int               ex    = EX_OK;

    ifBLIN_QX3("+ %08X %u", cf->flags, cf->num);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    convey = &cf->convey[cf->num];
    hlen = asprintf( &buf
                   , "P7\nWIDTH %u\nHEIGHT %u\nDEPTH 4\nMAXVAL 65535\nTUPLTYPE RGB_ALPHA\nENDHDR\n"
                   , (convey->flags & PNGTR_ROTSWAP) ? convey->IHDR.height : convey->IHDR.width
                   , (convey->flags & PNGTR_ROTSWAP) ? convey->IHDR.width : convey->IHDR.height
                   );
    pngnm = !convey->png.nm ? "" : convey->png.nm;
    PNGTR_W((hlen != mife_writ(fd, buf, hlen)),                     ("mife_writ %s", pngnm), EX_IOERR);
    free(buf);
    PNGTR_W(!(img = pngtr_dopam(cf, cf->num)),                   ("outpam %d", cf->num), EX_CANTCREAT);
    dopam = convey->IHDR.width * convey->IHDR.height * sizeof(pngtr_syncolor);
    PNGTR_W((dopam != mife_writ(fd, img, dopam)),                   ("mife_writ %s", pngnm), EX_IOERR);
    free(img);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef  blin_internal_flags
}

static char *
/*****************************************************************************************************
 ****                                                                                  ****
 *****************************************************************************************************
                                                                                                  ****/
tocopy(char *nm) {                                                                               /****
                                                                                                  ****
 *****************************************************************************************************/
    char           *buf  = NULL;

    if  (!!(buf = calloc(2 * strlen(nm) + 2, 1))) {
        for (int j = 0, k = 0; !!nm[j]; ++j) {
            switch(nm[j]) {
            case '\b': buf[k++] = '\\'; buf[k++] = 'b';
            case '\f': buf[k++] = '\\'; buf[k++] = 'f';
            case '\n': buf[k++] = '\\'; buf[k++] = 'n';
            case '\r': buf[k++] = '\\'; buf[k++] = 'r';
            case '\t': buf[k++] = '\\'; buf[k++] = 't';
            case '\v': buf[k++] = '\\'; buf[k++] = 'v';
            case '\\': buf[k++] = '\\'; buf[k++] = '\\';
            default  : buf[k++] = nm[j];
    }   }   }
    return(buf);
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****                                                                       ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_cmpam(pngtr_conf *cf, u_int i, BLIN_flag f, FILE *pr) {                                    /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey   *convey0;
    pngtr_convey   *convey1;
    u_int64_t       difp = 0;
    u_int64_t       difv = 0;
    pngtr_syncolor *img0 = NULL;
    pngtr_syncolor *img1 = NULL;
    pngtr_syncolor *pix0;
    pngtr_syncolor *pix1;
    char           *buf  = NULL;
    u_int64_t       vol;
    u_int           num  = cf->num;
    int             ex   = EX_OK;
    u_int           h0;
    u_int           w0;
    u_int           h1;
    u_int           w1;
    u_int           z    = 0;
    u_int           h;
    u_int           w;
    char           *c;

    PNGTR_X((num >= PNGTR_CONVEYNUM),                    ("Ill convey num %d", num), ERANGE, EX_USAGE);
    PNGTR_X((i >= PNGTR_CONVEYNUM),                        ("Ill convey num %d", i), ERANGE, EX_USAGE);
    convey0 = &cf->convey[cf->num];
    convey1 = &cf->convey[i];
    if  (convey0->flags & PNGTR_ROTSWAP) {
        w0 = convey0->IHDR.height;
        h0 = convey0->IHDR.width;
    } else {
        w0 = convey0->IHDR.width;
        h0 = convey0->IHDR.height;
    }
    if  (convey1->flags & PNGTR_ROTSWAP) {
        w1 = convey1->IHDR.height;
        h1 = convey1->IHDR.width;
    } else {
        w1 = convey1->IHDR.width;
        h1 = convey1->IHDR.height;
    }
    w = (w0 > w1) ? w0 : w1;
    h = (h0 > h1) ? h0 : h1;
    vol = (u_int64_t)h * w;
    if  (cf->num != i) {
        if  (!convey0->img.cont) {
            PNGTR_O( !!(ex = pngtr_flimg(cf, NULL))
                   , ("pngtr_flimg %s", !convey0->png.nm ? "" : convey0->png.nm)
                   );
        }
        cf->num = i;
        if  (!convey1->img.cont) {
            PNGTR_O( !!(ex = pngtr_flimg(cf, NULL))
                   , ("pngtr_flimg %s", !convey1->png.nm ? "" : convey1->png.nm)
                   );
        }
        cf->num = num;
        PNGTR_W(!(img0 = pngtr_dopam(cf, cf->num)),              ("outpam %d", cf->num), EX_CANTCREAT);
        PNGTR_W(!(img1 = pngtr_dopam(cf, i)),                          ("outpam %d", i), EX_CANTCREAT);
        for (u_int x = 0; x < w; ++x) {
            for (u_int y = 0; y < h; ++y) {
                pix0 = pix1 = NULL;
                if  ((x < w0) && (y < h0)) pix0 = &img0[coord(convey0, y, x)];
                if  ((x < w1) && (y < h1)) pix1 = &img1[coord(convey1, y, x)];
                if  (!pix0 && !pix1) {
                } else if (!!pix0 && !!pix1) {
                    z = 0;
                    z += (pix0->r > pix1->r) ? (pix0->r - pix1->r) : (pix1->r - pix0->r);
                    z += (pix0->g > pix1->g) ? (pix0->g - pix1->g) : (pix1->g - pix0->g);
                    z += (pix0->b > pix1->b) ? (pix0->b - pix1->b) : (pix1->b - pix0->b);
                    z += (pix0->a > pix1->a) ? (pix0->a - pix1->a) : (pix1->a - pix0->a);
                    if  (!!z) difp++;
                    difv += z;
                } else {
                    difp++;
                    difv += 262143;
        }   }   }
        free(img0);
        free(img1);
        img0 = NULL;
        img1 = NULL;
    }
out:
    c = "";
    if  (PNGTR_PRINTFN & f) {
        if  (!convey0->png.nm) {
            fprintf(pr, "");
        } else {
            fprintf(pr, "%s", (buf = tocopy(convey0->png.nm)));
            free(buf);
        }
        c = "\t";
    }
    if  (PNGTR_PRINTF2 & f) {
        if  (!convey1->png.nm) {
            fprintf(pr, "");
        } else {
            fprintf(pr, "%s%s", c, (buf = tocopy(convey1->png.nm)));
            free(buf);
        }
        c = "\t";
    }
    if  (f & PNGTR_DIFFCOD) {
        fprintf(pr, "%s%llu", c, (unsigned long long)vol);
        c = "\t";
    }
    if  (f & PNGTR_DIFFPNT) {
        fprintf(pr, "%s%llu", c, (unsigned long long)difp);
        c = "\t";
    }
    if  (f & PNGTR_DIFFVAL) {
        fprintf(pr, "%s%llu", c, (unsigned long long)((4 * vol + difv - 1) / vol / 4));
    }
    if  (!ex) {
        if  (!!difp) ex = 1;
        if  (!!difv) ex += 2;
    }
    if  (f & (PNGTR_DIFFCOD | PNGTR_DIFFPNT | PNGTR_DIFFVAL)) fprintf(pr, "\n");
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef  blin_internal_flags
}
